import importlib
import random
import time
from binascii import hexlify, unhexlify
from hashlib import blake2b

import cpuinfo

from .exceptions import InvalidDifficulty, InvalidMultiplier, InvalidWork
from .util import dec_to_hex, is_hex

# Select the PoW C extension depending on highest supported instruction set
# based on the following priorities:
# AVX > SSE4.1 > SSSE3 > SSE2 > reference implementation
#
# This based on a Ryzen 1800X giving the following results:
# avx speed: 6185344 hashes/s (this is likely the fastest on Intel CPUs)
# sse4_1 speed: 6259267 hashes/s
# ssse3 speed: 6249287 hashes/s
# sse2 speed: 4635929 hashes/s
# ref speed: 4539306 hashes/s

# TODO: Maybe run a short benchmark when running solve_work() for the first
#       time?
_cpu_flags = cpuinfo.get_cpu_info()["flags"]
_cpu_flags_by_priority = ("avx", "sse4_1", "ssse3", "sse2", "neon", "ref")

for cpu_flag in _cpu_flags_by_priority:
    if cpu_flag == "ref":
        _work = importlib.import_module("nanolib._work_ref")
        break
    elif cpu_flag in _cpu_flags:
        _work = importlib.import_module(
            "nanolib._work_{}".format(cpu_flag)
        )
        break


WORK_DIFFICULTY = "ffffffc000000000"
WORK_DIFFICULTY_INT = int(WORK_DIFFICULTY, 16)


__all__ = (
    "WORK_DIFFICULTY", "parse_work", "validate_work", "validate_difficulty",
    "derive_work_difficulty", "derive_work_multiplier",
    "get_work_value", "solve_work"
)


def parse_work(work):
    """
    Parses a proof-of-work (PoW) value and returns it if it's syntactically
    valid.

    .. note:: This method only checks that the work's format is correct.
              To validate a proof-of-work,
              use :func:`nanolib.work.validate_work`

    :param str work: Work as a 16-character hex string
    :raises InvalidWork: If the work is invalid
    :return: Work as a 16-character hex string
    :rtype: str
    """
    if not len(work) == 16 or not is_hex(work):
        raise InvalidWork("Work has to be a 16-character hexadecimal string")

    # NANO node represents PoW with lowercase hexadecimal letters for some
    # reason. Let's do the same here just to be sure.
    return work.lower()


def get_work_value(block_hash, work, as_hex=False):
    """
    Get the proof-of-work value. The work value must be equal to or higher than
    the work difficulty to be considered valid.

    :param str block_hash: Block hash as a 64-character hex string
    :param str work: Work as a 16-character hex string
    :param bool as_hex: Whether to return the work value as a hex string or an
                        integer.
                        If True, a 16-character hex string is returned.
                        If False, a 64-bit integer is returned.
                        A 64-bit integer is returned by default.
    :raises InvalidWork: If the work isn't a 16-character hex string
    :raises InvalidBlockHash: If the block hash isn't a 64-character hex string
    :return: A 16-character hex string or a 64-bit integer depending on
             `as_hex`
    :rtype: int or str
    """
    # Import is deferred to avoid a circular import
    from nanolib.blocks import validate_block_hash

    validate_block_hash(block_hash)
    parse_work(work)

    reversed_work = bytearray(unhexlify(work))
    reversed_work.reverse()
    work_hash = bytearray(blake2b(
        b"".join([reversed_work, unhexlify(block_hash)]),
        digest_size=8).digest())
    work_hash.reverse()
    work_value = int(hexlify(work_hash), 16)

    if as_hex:
        work_value = dec_to_hex(work_value, 8).lower()

    return work_value


def validate_work(block_hash, work, difficulty=WORK_DIFFICULTY):
    """Validate the proof-of-work.

    :param str block_hash: Block hash as a 64-character hex string
    :param str work: Work as a 16-character hex string
    :param str difficulty: The difficulty for the proof-of-work
                           as a 16-character hex string.
                           NANO network's difficulty is used by default.
    :raises InvalidWork: If the work doesn't meet the difficulty
    :return: The work as a 16-character hex string
    :rtype: str
    """
    difficulty = parse_difficulty(difficulty)

    work_value = get_work_value(block_hash=block_hash, work=work)

    if work_value < difficulty:
        raise InvalidWork("Work doesn't meet the required difficulty")

    return work.lower()


def validate_difficulty(difficulty):
    """Validate the work difficulty.

    :param str difficulty: Work difficulty as a 16-character hex string
    :raises InvalidDifficulty: If the difficulty isn't a 16-character hex value
    :return: The work difficulty as a 16-character hex string
    :rtype: str
    """
    if not len(difficulty) == 16 or not is_hex(difficulty):
        raise InvalidDifficulty(
            "Threshold has to be a 16-character hex string"
        )

    return difficulty.lower()


def parse_difficulty(difficulty):
    """Parse and return given hex-formatted difficulty as an integer.

    :param str difficulty: Work difficulty as a 16-character hex string
    :raises InvalidDifficulty: If the difficulty isn't a 16-character hex
                               string
    :return: The work difficulty as an integer
    :rtype: int
    """
    if is_hex(difficulty) and len(difficulty) == 16:
        return int(difficulty, 16)

    raise InvalidDifficulty("Difficulty has to be a 16-character hex string")


def derive_work_difficulty(multiplier, base_difficulty=None):
    """Derive the work difficulty from a provided multiplier
    and a base difficulty

    :param float multiplier: Work multiplier as a float. Work difficulty with
                             a multiplier of 2 requires on average twice as
                             much work compared to the base difficulty.
    :param str base_difficulty: Base difficulty as a 16-character hex string.
                                NANO network's difficulty is used by default.
    :raises InvalidDifficulty: If the difficulty isn't a 16-character hex
                               string
    :raises ValueError: If the resulting difficulty is larger than \
                        :math:`2^{64}-1`
    :return: The adjusted work difficulty as a 16-character hex string
    :rtype: str
    """
    if base_difficulty is None:
        base_difficulty = WORK_DIFFICULTY_INT
    else:
        base_difficulty = parse_difficulty(base_difficulty)

    try:
        multiplier = float(multiplier)
    except ValueError:
        raise InvalidMultiplier("Multiplier is not a float")

    if multiplier <= 0:
        raise InvalidMultiplier(
            "Multiplier has to be a positive non-zero float")

    difficulty = int((base_difficulty - (1 << 64)) / multiplier + (1 << 64))

    if difficulty > (2**64)-1:
        raise ValueError("Resulting difficulty is too large")

    return dec_to_hex(difficulty, 8).lower()


def derive_work_multiplier(difficulty, base_difficulty=None):
    """Derive the work multiplier from a difficulty and a base difficulty.

    :param str difficulty: Work difficulty as a 16-character hex string
    :param str base_difficulty: Base work difficulty as a 16-character hex
                                string.
                                NANO network's difficulty is used by default.
    :raises InvalidDifficulty: If difficulty isn't a 16-character hex string
    :return: The work multiplier as a float. Work difficulty with a multiplier
             of 2 requires on average twice as much work compared to the base
             difficulty.
    :rtype: float
    """
    if base_difficulty is None:
        base_difficulty = WORK_DIFFICULTY_INT
    else:
        base_difficulty = parse_difficulty(base_difficulty)

    difficulty = parse_difficulty(difficulty)

    multiplier = \
        float((1 << 64) - base_difficulty) / float((1 << 64) - difficulty)
    return multiplier


def solve_work(block_hash, difficulty=WORK_DIFFICULTY, timeout=None):
    """Solve the work for the corresponding block hash.

    :param str block_hash: Block hash as a 64-character hex string
    :param str difficulty: The difficulty for the proof-of-work as a
                           16-character hex string.
                           NANO network's difficulty is used by default.
    :param timeout: Timeout in seconds. If provided, None will be returned
                    if the work can't be solved in the given time.
                    If None, the function will block until the work is solved.
    :type timeout: int, float or None
    :return: The solved work as a 64-character hex string or None
             couldn't be solved in time
    :rtype: str or None
    """
    validate_difficulty(difficulty)

    # Reinitialize the random number generator in case this method is being
    # run in a multiprocessing environment; otherwise every process will
    # inherit the same RNG state
    random.seed()

    nonce = random.randint(0, (2**64)-1)
    block_hash_b = unhexlify(block_hash)

    start = time.time()

    while True:
        nonce = _work.do_work(
            block_hash_b, nonce, parse_difficulty(difficulty))
        work = hexlify(int(nonce).to_bytes(8, byteorder="big"))
        try:
            validate_work(block_hash, work, difficulty)
            return str(work, "utf-8")
        except InvalidWork:
            pass

        if timeout and (time.time() - start) > timeout:
            return None



"""
nanolib
~~~~~~~

nanolib is a Python library allowing to work with NANO cryptocurrency
functions such as block creation and manipulation, account generation and
proof-of-work validation and solving
"""
from .accounts import *
from .blocks import *
from .exceptions import *
from .units import *
from .work import *
from .util import *


from ._version import get_versions
__version__ = get_versions()['version']
del get_versions



"""
nanolib.blocks
~~~~~~~~~~~~~~

Methods to work with NANO blocks and a :class:`Block` class to construct
(either manually or from JSON) and process NANO blocks
"""
from base64 import b64decode, b64encode
from binascii import hexlify, unhexlify
from functools import wraps
from hashlib import blake2b
from json import dumps, loads

from ed25519_blake2b import BadSignatureError, SigningKey, VerifyingKey

from .accounts import (get_account_id, get_account_public_key,
                       validate_account_id, validate_private_key,
                       validate_public_key)
from .exceptions import (InvalidBalance, InvalidBlock, InvalidBlockHash,
                         InvalidSignature, InvalidWork)
from .util import dec_to_hex, is_hex
from .work import (WORK_DIFFICULTY, get_work_value, parse_difficulty,
                   solve_work, validate_difficulty, validate_work)

__all__ = (
    "balance_to_hex", "parse_hex_balance", "parse_signature",
    "validate_balance", "validate_block_hash", "Block"
)


BLOCK_TYPES = ("send", "receive", "open", "change", "state")
BLOCK_PARAMS = (
    "type", "account", "previous", "destination", "representative",
    "balance", "source", "link", "link_as_account", "signature", "work"
)
BLOCK_REQUIRED_PARAMS = {
    "send": ("type", "previous", "destination", "balance"),
    "receive": ("type", "previous", "source"),
    "open": ("type", "source", "representative", "account"),
    "change": ("type", "previous", "representative"),
    "state": (
        "type", "account", "previous", "representative", "balance", "link"
    ),
}
BLOCK_OPTIONAL_PARAMS = {
    "send": ("work", "signature", "account"),
    "receive": ("work", "signature", "account"),
    "open": ("work", "signature"),
    "change": ("work", "signature", "account"),
    "state": ("work", "signature", "link_as_account")
}

# Block hash of all zeroes is used as an empty value for the 'state' field
ZERO_BLOCK_HASH = "0"*64
ZERO_ACCOUNT_ID = \
    "xrb_1111111111111111111111111111111111111111111111111111hifc8npp"


STATE_BLOCK_HEADER_BYTES = unhexlify(
    "0000000000000000000000000000000000000000000000000000000000000006")

# Constants used for epoch blocks. For details, see:
# https://github.com/nanocurrency/nano-node/pull/955
#
# 'link' value used to denote an account upgrade from V0 to V1
# The resulting bytearray starts with the string 'epoch v1 block'
EPOCH_LINK_V1 = \
    "65706F636820763120626C6F636B000000000000000000000000000000000000"
EPOCH_SIGN_PUBLIC_KEY = \
    "e89208dd038fbb269987689621d52292ae9c35941a7484756ecced92a65093ba"


MAX_BALANCE = 2**128 - 1


def block_parameter(setter):
    """Raise an error if a parameter is set to an invalid value
    """
    @wraps(setter)
    def wrapper(self, new):
        parameter = setter.__name__[4:]
        try:
            setter(self, new)
        except ValueError as e:
            raise type(e)(
                "Encountered exception while changing block parameter "
                "'{parameter}'".format(
                    parameter=parameter,
                    message=str(e))
            ) from e

    return wrapper


def invalidate_signature(setter):
    """
    Invalidate the cached value for `has_valid_signature` when the setter
    is called
    """
    @wraps(setter)
    def wrapper(self, val):
        self._has_valid_signature = None
        setter(self, val)

    return wrapper


def invalidate_work(setter):
    """
    Invalidate the cached value for `has_valid_work` when the setter
    is called
    """
    @wraps(setter)
    def wrapper(self, val):
        self._has_valid_work = None
        setter(self, val)

    return wrapper


def balance_to_hex(balance):
    """Convert a NANO balance to a 16-character hex string used in
    serialized legacy send blocks

    :param balance: Balance to convert
    :return: Balance as a 16-character hex string with necessary padding
    :rtype: str
    """
    return dec_to_hex(balance, 16)


def parse_hex_balance(balance):
    """Parse and return given hex-formatted balance as an integer.
    Hex-formatted balance is used in legacy send blocks

    :param str balance: Hex-formatted balance to convert
    :return: Converted balance
    :rtype: int
    """
    # Balance can also be given as a hexadecimal value
    if is_hex(balance) and len(balance) == 32:
        return int(balance, 16)

    raise InvalidBalance(
        "Balance for legacy send block needs to be a hexadecimal string")


def parse_signature(signature):
    """Parse a signature and return it if it's syntactically valid

    .. note:: This method only checks that the signature's format is correct.
              To verify a signature, create a :class:`Block` and use its
              :meth:`Block.verify_signature`

    :param str signature: Signature as a 128-character hex string
    :raises ValueError: Signature is invalid
    :return: Signature in uppercase
    :rtype: str
    """
    if not len(signature) == 128 or not is_hex(signature):
        raise InvalidSignature(
            "Signature has to be a 128-character hexadecimal string")

    return signature.upper()


def validate_balance(balance):
    """Validate the balance

    :param int balance: Balance
    :raises InvalidBalance: If the balance is less than 0 or higher than the
                            maximum possible balance (2^128 - 1)
    :return: Balance
    :rtype: int
    """
    if not isinstance(balance, int) or (balance < 0 or balance > MAX_BALANCE):
        raise InvalidBalance(
            "Balance has to be an integer in range 0 - (2^128 - 1)")

    return balance


def validate_block_hash(h):
    """Validate the block hash

    :param str h: Block hash as a 64-character hex string
    :raises InvalidBlockHash: If the block hash is invalid
    :return: Block hash in uppercase
    :rtype: str
    """
    if not len(h) == 64 or not is_hex(h):
        raise InvalidBlockHash(
            "Block hash has to be a 64-character hexadecimal string")

    return h.upper()


class Block(object):
    """A NANO block.

    Can be constructed manually or from a JSON string

    .. note:: For deserializing existing blocks, see \
              :meth:`Block.from_json` and :meth:`Block.from_dict`
    """
    __slots__ = (
        "_block_type", "_account", "_previous", "_destination",
        "_representative", "_balance", "_source", "_link", "_signature",
        "_work", "_difficulty", "_has_valid_signature", "_has_valid_work"
    )

    def __init__(self, block_type, verify=True, difficulty=None, **kwargs):
        """Create a block from given parameters

        .. note:: Since `type` is a Python keyword, :ivar:`block_type` is \
                  used to reference the `type` field used in NANO blocks


        :param str block_type: The type of block. Can be either
                               `send`, `receive`, `open`, `change` or `state`
        :param bool verify: If True, signature and/or the proof-of-work will
                            be verified
        :param str difficulty: Work difficulty used to check PoW's validity.
                               Default is NANO main net's default work
                               difficulty.
        :raises InvalidBlock: If the required arguments weren't provided for
                              the block type, or prohibited arguments
                              were provided
        :raises InvalidSignature: If the signature was provided but was
                                  found to be invalid
        :raises InvalidWork: If work was provided but was found to be invalid
        :raises InvalidDifficulty: If invalid work difficulty was provided
        :return: The created block
        :rtype: Block
        """
        self._has_valid_signature = None
        self._has_valid_work = None
        self.block_type = block_type

        # Set None as default value for all parameters except block_type
        for key in Block.__slots__[1:]:
            setattr(self, key, None)

        for key, value in kwargs.items():
            setattr(self, key, value)

        if not difficulty:
            difficulty = WORK_DIFFICULTY
        self.difficulty = difficulty

        self._validate(verify=verify)

    def verify_work(self, difficulty=None):
        """Verify the work in the block

        :param str difficulty: The difficulty for the proof-of-work.
                               NANO mainnet difficulty is used by default.
        :raises ValueError: If work isn't included in the block
        :raises InvalidWork: If included work doesn't meet the difficulty
        """
        if not difficulty:
            difficulty = self.difficulty

        if not self.work:
            raise ValueError("Work hasn't been added to this block")

        validate_work(
            block_hash=self.work_block_hash, work=self.work,
            difficulty=difficulty)

    def verify_signature(self):
        """Verify the signature in the block

        .. note:: :attr:`Block.account` and :attr:`Block.signature` need to \
                  be set to verify the signature

        :raises ValueError: :attr:`Block.account` or :attr:`Block.signature` \
                            isn't set
        :raises InvalidSignature: Signature was verified and found to be \
                                  invalid
        """
        if not self.account:
            raise ValueError(
                "'account' not included for signature verification. "
                "To verify the signature, the block needs to have a valid "
                "'account' value."
            )

        if not self.signature:
            raise ValueError("Signature hasn't been added to this block")

        if self.tx_type == "epoch":
            # Epoch blocks are signed by the genesis account regardless
            # of the actual account
            public_key = EPOCH_SIGN_PUBLIC_KEY
        else:
            public_key = get_account_public_key(account_id=self.account)
            validate_public_key(public_key)

        vk = VerifyingKey(unhexlify(public_key))

        try:
            vk.verify(
                sig=unhexlify(self.signature),
                msg=unhexlify(self.block_hash)
            )
        except BadSignatureError:
            raise InvalidSignature("Signature couldn't be verified")

    def sign(self, private_key):
        """Sign the block and set the value for :attr:`Block.signature`

        :raises ValueError: If the block already has a signature
        :return: True if the signature was added successfully
        :rtype: bool
        """
        if self.has_valid_signature:
            raise ValueError("The block already has a signature.")

        validate_private_key(private_key)

        sk = SigningKey(unhexlify(private_key))

        sig = sk.sign(msg=unhexlify(self.block_hash))
        self.signature = hexlify(sig).decode()

        return True

    def solve_work(self, difficulty=None, timeout=None):
        """Solve the work contained in this block and update the Block
        instance to include the work

        :raises ValueError: If the block already has valid proof-of-work
                            meeting the difficulty
        :param str difficulty: The difficulty for the proof-of-work.
                               NANO mainnet difficulty is used by default.
        :param timeout: Timeout in seconds. If provided, None will be returned
                        if the work can't be solved in the given time.
                        If None, the function will block until the work is solved.
        :type timeout: int, float or None
        :return: True if the work was solved in the given time, False otherwise
        :rtype: bool
        """
        if not difficulty:
            difficulty = self.difficulty
        else:
            self.difficulty = validate_difficulty(difficulty)

        if self.work:
            try:
                self.verify_work(difficulty=difficulty)
                raise ValueError("Block already has a valid proof-of-work")
            except InvalidWork:
                pass

        result = solve_work(
            block_hash=self.work_block_hash, difficulty=difficulty,
            timeout=timeout)

        if result:
            self.work = result
            return True

        return False

    def _validate(self, verify=True):
        """
        Make sure the block has correct parameters corresponding to its type

        .. note:: This is called automatically when creating the Block instance.
                  You don't need to call this manually.

        :param bool verify: If True, signature and/or proof-of-work will
                            be verified if included in the block
        :raises InvalidBlock: If the block has missing or prohibited parameters
        :raises InvalidSignature: If the signature was provided and was found
                                  to be invalid
        :raises InvalidWork: If work was provided and was found to be below
                             the required difficulty
        """
        block_params = set(self.to_dict().keys())
        required_params = set(BLOCK_REQUIRED_PARAMS[self.block_type])
        optional_params = set(BLOCK_OPTIONAL_PARAMS[self.block_type])

        # Check for missing params
        if len(required_params - block_params) != 0:
            missing_params = required_params - block_params
            raise InvalidBlock(
                "Block with type '{block_type}' is missing required "
                "parameters: {missing}".format(
                    block_type=self.block_type,
                    missing=", ".join(missing_params))
            )

        # Check for invalid params, excluding work and signature which are
        # optional
        if len(block_params - required_params - optional_params) != 0:
            prohibited_params = (
                block_params - required_params - optional_params)
            raise InvalidBlock(
                "Block with type '{block_type}' has prohibited parameters: "
                "{prohibited}".format(
                    block_type=self.block_type,
                    prohibited=", ".join(prohibited_params))
            )

        if verify:
            # If signature and account is included, verify the signature
            if self.signature and self.account:
                self.verify_signature()

            # If work is included, verify it
            if self.work:
                self.verify_work()

    def to_dict(self):
        """Return a :type:`dict` consisting of items used in the block
        that can be broadcast to the NANO network

        .. note:: The :ivar:`Block.block_type` has the key name `type`
                  in the created :type:`dict`

        :return: A dictionary of block items
        :rtype: dict
        """
        block_items = {}

        for param in BLOCK_PARAMS:
            if param == "type":
                param = "block_type"

            val = getattr(self, param)

            if val is not None:
                block_items[param] = val

        if block_items["block_type"]:
            block_items["type"] = block_items["block_type"]
            del block_items["block_type"]

        if "balance" in block_items.keys():
            if self.block_type == "state":
                block_items["balance"] = str(block_items["balance"])
            else:
                block_items["balance"] = balance_to_hex(
                    block_items["balance"])

        return block_items

    def json(self):
        """Return a JSON-formatted string of the block that can be
        broadcast to the NANO network

        :return: A JSON-formatted string
        :rtype: str
        """
        block_items = self.to_dict()

        return dumps(block_items)

    @classmethod
    def from_json(cls, json, verify=True, difficulty=None):
        """Create a :class:`Block` instance from a JSON-formatted string

        :param str json: A JSON-formatted block to deserialize
        :param bool verify: If True, signature and/or the proof-of-work will
                            be verified
        :param str difficulty: Work difficulty used to check PoW's validity.
                              Default is NANO main net's default work
                              difficulty.
        :return: Block
        :rtype: Block
        """
        block_items = loads(json)

        return cls.from_dict(block_items, verify=verify, difficulty=difficulty)

    @classmethod
    def from_dict(cls, d, verify=True, difficulty=None):
        """Create a :class:`Block` instance from a dictionary

        :param dict d: The block fields to deserialize
        :param bool verify: If True, signature and/or the proof-of-work will
                            be verified
        :param str difficulty: Work difficulty used to check PoW's validity.
                              Default is NANO main net's default work
                              difficulty.
        :return: Block
        :rtype: Block
        """
        d = d.copy()

        d["block_type"] = d["type"]
        del d["type"]

        if "balance" in d.keys():
            # Blocks that predate state blocks use hexadecimal balances
            if d["block_type"] == "send":
                d["balance"] = parse_hex_balance(d["balance"])
            else:
                d["balance"] = int(d["balance"])

        return cls(**d, verify=verify, difficulty=difficulty)

    @property
    def tx_type(self):
        """Return the transaction type of this block

        For blocks besides 'state', this is the same as 'block_type'. For
        'state', the type is derived from the value of 'link' field

        :return: For legacy blocks, the transaction type can be \
                 `open`, `change`, `receive` or `send`. \
                 For state blocks, the transaction type can be \
                 `change`, `open`, `send/receive` or `epoch`
        """
        if self.block_type != "state":
            return self.block_type
        elif self.link == ZERO_BLOCK_HASH:
            return "change"
        elif self.link == EPOCH_LINK_V1:
            return "epoch"
        elif self.previous == ZERO_BLOCK_HASH:
            return "open"

        return "send/receive"

    @property
    def has_valid_signature(self):
        """Check if the block has a valid signature

        :return: True if the block has a valid signature, \
                 False if either :attr:`Block.signature` or \
                 :attr:`Block.account` is missing, or if the signature \
                 was found to be invalid
        :rtype: bool
        """
        if self._has_valid_signature is not None:
            return self._has_valid_signature

        if not self.signature or not self.account:
            self._has_valid_signature = False
            return False

        try:
            self.verify_signature()
            self._has_valid_signature = True
        except InvalidSignature:
            self._has_valid_signature = False

        return self._has_valid_signature

    @property
    def has_valid_work(self):
        """Check if the block has valid work.

        .. note::

           This method assumes that NANO mainnet difficulty is used. In any
           other case use :meth:`nanolib.blocks.Block.verify_work`
           instead.

        :return: True if the block has valid work that meets the difficulty, \
                 False if :attr:`Block.work` is missing or the work
                 was found to be below the required difficulty
        :rtype: bool
        """
        if self._has_valid_work is not None:
            return self._has_valid_work

        if not self.work:
            self._has_valid_work = False
            return False

        try:
            self.verify_work()
            self._has_valid_work = True
        except InvalidWork:
            self._has_valid_work = False

        return self._has_valid_work

    @property
    def complete(self):
        """Check if the block has a valid signature and valid work

        :return: True if both the signature and work are included and both \
                 were verified to be correct, False otherwise
        :rtype: bool
        """
        return self.has_valid_signature and self.has_valid_work

    @property
    def work_block_hash(self):
        """BLAKE2b hash that requires a valid PoW

        For open blocks this is the public key derived from
        :attr:`nanolib.blocks.Block.account`. For other types of blocks,
        :attr:`nanolib.blocks.Block.previous` is used as the hash.

        :return: BLAKE2b hash used to generate a PoW
        :rtype: str

        """
        # 'open' blocks use the account ID ('account') itself as part of the
        # hash
        # Other blocks use 'previous' to make sure PoW can be completed
        # in advance (user can immediately send the transaction since PoW is
        # already ready)
        return (
            get_account_public_key(account_id=self.account)
            if self.tx_type == "open"
            else self.previous
        )

    @property
    def work_value(self):
        """The work value attached to this block. The value must be equal to
        or higher than :attr:`nanolib.blocks.Block.difficulty`
        in order to be valid.

        :return: 64-bit integer or None if this block doesn't include work
        :rtype: int or None
        """
        if self.work:
            return get_work_value(
                block_hash=self.work_block_hash, work=self.work)

        return None

    @property
    def block_hash(self):
        """BLAKE2b hash for the block used to identify the block

        :raises InvalidBlock: If the block hash can't be calculated for the \
                              block
        :return: Block hash as a 64-character hex string
        :rtype: str
        """
        if self.block_type == "receive":
            return blake2b(
                b"".join([
                    self._previous,
                    self._source
                ]),
                digest_size=32
            ).hexdigest().upper()
        elif self.block_type == "open":
            return blake2b(
                b"".join([
                    self._source,
                    unhexlify(
                        get_account_public_key(account_id=self.representative)
                    ),
                    unhexlify(
                        get_account_public_key(account_id=self.account)
                    )
                ]),
                digest_size=32
            ).hexdigest().upper()
        elif self.block_type == "change":
            return blake2b(
                b"".join([
                    self._previous,
                    unhexlify(
                        get_account_public_key(account_id=self.representative)
                    )
                ]),
                digest_size=32
            ).hexdigest().upper()
        elif self.block_type == "send":
            return blake2b(
                b"".join([
                    self._previous,
                    unhexlify(
                        get_account_public_key(account_id=self.destination)
                    ),
                    unhexlify(balance_to_hex(self.balance))
                ]),
                digest_size=32
            ).hexdigest().upper()
        elif self.block_type == "state":
            data = b"".join([
                STATE_BLOCK_HEADER_BYTES,
                unhexlify(
                    get_account_public_key(account_id=self.account)
                ),
                self._previous,
                unhexlify(
                    get_account_public_key(account_id=self.representative)
                ),
                unhexlify(balance_to_hex(self.balance)),
                self._link
            ])
            return blake2b(data, digest_size=32).hexdigest().upper()
        else:
            raise InvalidBlock(
                "Block hash can't be calculated for type {block_type}".format(
                    block_type=self.block_type)
            )

    @block_parameter
    @invalidate_signature
    def set_block_type(self, block_type):
        if block_type not in BLOCK_TYPES:
            raise ValueError("Block type is not valid")

        self._block_type = block_type

    @block_parameter
    @invalidate_signature
    @invalidate_work
    def set_account(self, account):
        if account is not None:
            validate_account_id(account)
        self._account = account

    @block_parameter
    @invalidate_signature
    @invalidate_work
    def set_source(self, source):
        if source is not None:
            validate_block_hash(source)
            self._source = unhexlify(source.encode("utf-8"))
        else:
            self._source = None

    def get_source(self):
        if not self._source:
            return None

        return hexlify(self._source).decode("utf-8").upper()

    @block_parameter
    @invalidate_signature
    @invalidate_work
    def set_previous(self, previous):
        if previous is not None:
            validate_public_key(previous)
        else:
            previous = ZERO_BLOCK_HASH

        self._previous = unhexlify(previous.encode("utf-8"))

    def get_previous(self):
        if not self._previous:
            return None

        return hexlify(self._previous).decode("utf-8").upper()

    @block_parameter
    @invalidate_signature
    def set_destination(self, destination):
        if destination is not None:
            validate_account_id(destination)
        self._destination = destination

    @block_parameter
    @invalidate_signature
    def set_representative(self, representative):
        if representative is not None:
            validate_account_id(representative)
        self._representative = representative

    @block_parameter
    @invalidate_signature
    def set_balance(self, balance):
        if balance is not None:
            validate_balance(balance)
        self._balance = balance

    @block_parameter
    @invalidate_signature
    def set_link(self, link):
        if link is not None:
            self._link = unhexlify(validate_block_hash(link).encode("utf-8"))
        else:
            self._link = unhexlify(ZERO_BLOCK_HASH.encode("utf-8"))

    def get_link(self):
        if not self._link:
            return None

        return hexlify(self._link).decode("utf-8").upper()

    @block_parameter
    @invalidate_signature
    def set_link_as_account(self, link_as_account):
        if link_as_account is not None:
            self._link = unhexlify(
                get_account_public_key(
                    account_id=link_as_account
                ).encode("utf-8")
            )
        else:
            self._link = unhexlify(ZERO_BLOCK_HASH.encode("utf-8"))

    def get_link_as_account(self):
        if not self._link:
            return None

        return get_account_id(public_key=self.link)

    @block_parameter
    @invalidate_signature
    def set_signature(self, signature):
        if signature is not None:
            self._signature = unhexlify(
                parse_signature(signature).encode("utf-8")
            )
        else:
            self._signature = None

    def get_signature(self):
        if not self._signature:
            return None

        return hexlify(self._signature).decode("utf-8").upper()

    @block_parameter
    @invalidate_work
    def set_work(self, work):
        if work is not None:
            self._work = unhexlify(work.encode("utf-8"))
        else:
            self._work = None

    def get_work(self):
        if not self._work:
            return None

        return hexlify(self._work).decode("utf-8")

    @invalidate_work
    def set_difficulty(self, difficulty):
        if difficulty is None:
            raise ValueError("'difficulty' property is required")

        self._difficulty = parse_difficulty(difficulty)

    block_type = property(lambda x: x._block_type, set_block_type)
    account = property(lambda x: x._account, set_account)
    source = property(get_source, set_source)
    previous = property(get_previous, set_previous)
    destination = property(lambda x: x._destination, set_destination)
    representative = property(lambda x: x._representative, set_representative)
    balance = property(lambda x: x._balance, set_balance)
    link = property(get_link, set_link)
    link_as_account = property(get_link_as_account, set_link_as_account)
    signature = property(get_signature, set_signature)
    work = property(get_work, set_work)
    difficulty = property(
        lambda x: dec_to_hex(x._difficulty, 8).lower(), set_difficulty
    )



"""
nanolib.util
~~~~~~~~~~~~

Functions for working with the variant of Base32 encoding
(referred to as "Nano Base32") used in account IDs, as well as other general
functions
"""
import nanolib._nbase32

__all__ = (
    "dec_to_hex", "is_hex", "nbase32_to_bytes", "bytes_to_nbase32"
)


def dec_to_hex(d, n):
    return format(d, "0{}X".format(n*2))
    # return "%0.{bytes}X".format(bytes=n*2) % d


def nbase32_to_bytes(nbase32):
    """
    Decode a Nano Base32 encoded string into bytes

    :param str nbase32: Nano Base32 encoded string
    :return: Decoded bytes
    :rtype: bytes
    """
    return nanolib._nbase32.nbase32_to_bytes(bytes(nbase32, "utf-8"))


def bytes_to_nbase32(b):
    """
    Encode bytes to Nano Base32

    :param bytes b: Bytes to encode
    :return: Encoded Nano Base32 string
    :rtype: str
    """
    return nanolib._nbase32.bytes_to_nbase32(b).decode("utf-8")


def is_hex(h):
    try:
        int(h, 16)
        return True
    except ValueError:
        return False



from ed25519_blake2b import BadSignatureError


__all__ = (
    "InvalidBlock", "InvalidWork", "InvalidDifficulty", "InvalidMultiplier",
    "InvalidSignature", "InvalidBlockHash", "InvalidBalance", "InvalidSeed",
    "InvalidAccount", "InvalidPrivateKey", "InvalidPublicKey"
)


class InvalidBlock(ValueError):
    """The block is invalid."""


class InvalidWork(ValueError):
    """The given work is invalid."""


class InvalidDifficulty(ValueError):
    """The given work difficulty is invalid."""


class InvalidMultiplier(ValueError):
    """The given work multiplier is invalid."""


class InvalidSignature(BadSignatureError):
    """The given signature is invalid."""


class InvalidBlockHash(ValueError):
    """The given block hash is invalid."""


class InvalidBalance(ValueError):
    """The given balance is invalid."""


class InvalidSeed(ValueError):
    """The seed is invalid."""


class InvalidAccount(ValueError):
    """The NANO account ID is invalid."""


class InvalidPrivateKey(ValueError):
    """The NANO private key is invalid."""


class InvalidPublicKey(ValueError):
    """The NANO public key is invalid."""



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Note: To use the 'upload' functionality of this file, you must:
#   $ pip install twine

import glob
import io
import os
import platform
import sys
from distutils.ccompiler import get_default_compiler
from shutil import rmtree

from setuptools import Command, Extension, setup
from setuptools.command.test import test as TestCommand

import versioneer

# Package meta-data.
NAME = 'nanolib'
DESCRIPTION = \
    'Python library for working with the NANO cryptocurrency protocol'
URL = 'https://github.com/Matoking/nanolib'
EMAIL = 'jannepulk@gmail.com'
AUTHOR = 'Janne Pulkkinen'
REQUIRES_PYTHON = '>=3.6.0'

# What packages are required for this module to be executed?
REQUIRED = [
    'ed25519-blake2b>=1.4', 'py-cpuinfo>=4'
]


def get_compile_args(iset=None):
    FLAGS = {
        "unix": {
            "avx": ["-DWORK_AVX", "-mavx"],
            "sse4_1": ["-DWORK_SSE4_1", "-msse4.1"],
            "ssse3": ["-DWORK_SSSE3", "-mssse3"],
            "sse2": ["-DWORK_SSE2", "-msse2"],
            "neon": ["-DWORK_NEON", "-mfpu=neon"],
            None: ["-DWORK_REF"]
        },
        "msvc": {
            "avx": ["/DWORK_AVX", "/arch:AVX", "/DHAVE_AVX", "/D__SSE4_1__"],
            "sse4_1": ["/DWORK_SSE4_1", "/arch:SSE2", "/D__SSE4_1__"],
            "ssse3": ["/DWORK_SSSE3", "/arch:SSE2", "/D__SSSE3__"],
            "sse2": ["/DWORK_SSE2", "/arch:SSE2", "/D__SSE2__"],
            "neon": ["/DWORK_NEON"],
            None: ["/DWORK_REF"]
        }
    }

    compiler = get_default_compiler()

    try:
        return FLAGS[compiler][iset]
    except KeyError:
        raise OSError("Compiler '{}' not supported.".format(compiler))


SOURCE_ROOT = os.path.join("src", "nanolib-work-module", "BLAKE2")
SOURCE_FILES = {
    "ref": glob.glob(os.path.join(SOURCE_ROOT, "ref", "blake2b*.c")),
    "sse": glob.glob(os.path.join(SOURCE_ROOT, "sse", "blake2b*.c")),
    "neon": glob.glob(os.path.join(SOURCE_ROOT, "neon", "blake2b-*.c"))
}


def create_work_extension(source_name="ref", iset=None):
    source_path = os.path.join(
        "src", "nanolib-work-module", "BLAKE2", source_name
    )
    module_suffix = iset if iset else "ref"

    return Extension(
        "nanolib._work_{}".format(module_suffix),
        include_dirs=[source_path],
        sources=[
            os.path.join("src", "nanolib-work-module", "work.c")
        ] + SOURCE_FILES[source_name],
        extra_compile_args=get_compile_args(iset)
    )


EXTENSIONS_TO_BUILD = []

_machine = platform.machine()

# https://stackoverflow.com/a/45125525
_is_arm = _machine.startswith("arm") or _machine.startswith("aarch64")
# 'AMD64' only appears on Windows
_is_x86 = _machine.startswith("x86") or _machine in ("i386", "i686", "AMD64")


if _is_x86:
    EXTENSIONS_TO_BUILD = [
        create_work_extension("sse", "avx"),
        create_work_extension("sse", "sse4_1"),
        create_work_extension("sse", "ssse3"),
        create_work_extension("sse", "sse2"),
        create_work_extension("ref")
    ]
elif _is_arm:
    EXTENSIONS_TO_BUILD = [
        create_work_extension("neon", "neon"),
        create_work_extension("ref")
    ]
else:
    EXTENSIONS_TO_BUILD = [create_work_extension("ref")]

EXTENSIONS_TO_BUILD.append(
    Extension(
        "nanolib._nbase32",
        include_dirs=[os.path.join("src", "nanolib-nbase32-module")],
        sources=[
            os.path.join("src", "nanolib-nbase32-module", "nbase32.c"),
            os.path.join("src", "nanolib-nbase32-module", "bit_array.c")
        ]
    )
)

# The rest you shouldn't have to touch too much :)
# ------------------------------------------------
# Except, perhaps the License and Trove Classifiers!
# If you do change the License, remember to change the Trove Classifier for that!

here = os.path.abspath(os.path.dirname(__file__))

# Import the README and use it as the long-description.
# Note: this will only work if 'README.rst' is present in your MANIFEST.in file!
with io.open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = '\n' + f.read()


class PyTest(TestCommand):
    user_options = [('pytest-args=', 'a', "Arguments to pass into py.test")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = ""

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        import shlex
        import pytest

        errno = pytest.main(shlex.split(self.pytest_args))
        sys.exit(errno)


class SpeedTest(Command):
    description = (
        "Run a set of speed tests to measure performance of cryptographic "
        "operations"
    )
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        from tests.performance_test import run_speed_tests

        run_speed_tests()


class UploadCommand(Command):
    """Support setup.py upload."""

    description = 'Build and publish the package.'
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print('\033[1m{0}\033[0m'.format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status('Removing previous builds…')
            rmtree(os.path.join(here, 'dist'))
        except OSError:
            pass

        self.status('Building Source and Wheel (universal) distribution…')
        os.system('{0} setup.py sdist bdist_wheel --universal'.format(sys.executable))

        self.status('Uploading the package to PyPi via Twine…')
        os.system('twine upload dist/*')

        self.status('Pushing git tags…')
        os.system('git tag v{0}'.format(versioneer.get_version()))
        os.system('git push --tags')

        sys.exit()


cmdclass = versioneer.get_cmdclass()
cmdclass.update({
    'upload': UploadCommand,
    'speed': SpeedTest,
    'pytest': PyTest,
})

# Where the magic happens:
setup(
    name=NAME,
    version=versioneer.get_version(),
    description=DESCRIPTION,
    long_description=long_description,
    author=AUTHOR,
    author_email=EMAIL,
    python_requires=REQUIRES_PYTHON,
    url=URL,
    ext_modules=EXTENSIONS_TO_BUILD,
    packages=["nanolib"],
    package_data={"": ["LICENSE"]},
    package_dir={"nanolib": "src/nanolib"},
    install_requires=REQUIRED,
    setup_requires=["sphinx"],
    tests_require=["pytest"],
    include_package_data=True,
    license='CC0',
    classifiers=[
        # Trove classifiers
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        #  'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: Implementation :: CPython',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX :: Linux',
        'Topic :: Office/Business :: Financial',
    ],
    # $ setup.py publish support.
    cmdclass=cmdclass,
    command_options={
        'build_sphinx': {
            'project': ('setup.py', NAME),
            'version': ('setup.py', versioneer.get_version()),
            'source_dir': ('setup.py', 'docs')
        },
    }
)



"""
nanolib.units
~~~~~~~~~~~~~

Methods for converting between different NANO denominations.

The smallest possible denomination is known as `raw`. The other
common denominations are:

* 1 nano = :math:`10^{24}` raw
* 1 knano = :math:`10^{27}` raw
* 1 Mnano/NANO = :math:`10^{30}` raw


Or in other words:

* 1 NANO = 1 000 000 nano
* 1 000 knano = 1 NANO
* 1 Mnano = 1 000 knano
* 1 Mnano = 1 000 000 nano
* 1 Mnano = 1 NANO

.. warning::
    Note that `nano` and `NANO` are **not** interchangeable!
    If you are using the denominations in a context where they may cause
    confusion, consider using `nano` and `Mnano` instead.

.. note::
    Denominations `XRB`/`Mrai`, `Krai` and `xrb`/`rai` were used before
    RaiBlocks' rebranding as NANO. The corresponding old names for
    these denominations are:

    * Mnano/NANO = Mrai/XRB
    * knano = kRAI
    * Mnano = Mrai
    * nano = xrb/rai

.. note::
    Consider storing all NANO values as integers using the `raw` denomination
    and only converting the value to another denomination
    when presenting a human-readable amount.
    `nanolib` uses Python's Decimal library and increases
    precision on startup to account for precision loss.
    This precision can be changed during runtime by another library.

"""
import functools
import enum
from decimal import Decimal, Inexact, getcontext

# Increase precision for Decimals in order to make it suitable for
# handling NANO denominations
getcontext().prec = 39
getcontext().traps[Inexact] = 1


__all__ = (
    "convert", "NanoDenomination", "NANO_RAW_CAP",
)


NANO_RAW_CAP = (2**128)-1


class NanoDenomination(enum.Enum):
    """Denominations used in NANO"""

    RAW = "raw"
    """Smallest division used in NANO"""

    MICRONANO = "unano"
    """1 unano = :math:`10^{18}` `raw`"""

    MILLINANO = "millinano"
    """1 mnano = :math:`10^{21}` `raw`"""

    NANO = "nano"
    """1 nano = :math:`10^{24}` `raw`"""

    KILONANO = "knano"
    """1 knano = :math:`10^{27}` `raw`"""

    MEGANANO = "Mnano"
    """1 Mnano = :math:`10^{30}` `raw`"""

    GIGANANO = "Gnano"
    """1 Gnano = :math:`10^{33}` `raw`"""


NANO_RAW_AMOUNTS = {
    NanoDenomination.RAW.value: Decimal("1"),
    NanoDenomination.MICRONANO.value: Decimal("1000000000000000000"),
    NanoDenomination.MILLINANO.value: Decimal("1000000000000000000000"),
    NanoDenomination.NANO.value: Decimal("1000000000000000000000000"),
    NanoDenomination.KILONANO.value: Decimal("1000000000000000000000000000"),
    NanoDenomination.MEGANANO.value:
        Decimal("1000000000000000000000000000000"),
    NanoDenomination.GIGANANO.value:
        Decimal("1000000000000000000000000000000000"),
}


def _forbid_float(func):
    """Decorator that only allows an integer or a Decimal as the first argument
    """
    @functools.wraps(func)
    def _decorator(*args, **kwargs):
        if not isinstance(args[0], int) and not isinstance(args[0], Decimal):
            raise TypeError(
                "Value has to be an integer or a Decimal. "
                "Floating numbers cause precision loss and are unsuitable "
                "for handling monetary values."
            )

        return func(*args, **kwargs)

    return _decorator


@_forbid_float
def convert(amount, source, target):
    """Convert an amount from one denomination to another

    :param amount: Amount to convert
    :type amount: decimal.Decimal or int
    :param NanoDenomination source: The denomination to convert from
    :param NanoDenomination target: The denomination to convert to
    :raises ValueError: If the amount is higher than the NANO coin cap \
                        (:math:`2^{128} - 1` `raw`)
    :raises TypeError": If `amount`  is not an `int` or a
                        :class:`decimal.Decimal`

    :return: Converted amount
    :rtype: decimal.Decimal
    """
    source = NanoDenomination(source)
    target = NanoDenomination(target)

    raw_source = NANO_RAW_AMOUNTS[source.value]
    raw_target = NANO_RAW_AMOUNTS[target.value]

    # Convert first into raw, then into the target denomination
    raw_amount = amount * raw_source

    if raw_amount > NANO_RAW_CAP:
        raise ValueError("Amount is higher than the NANO coin supply")

    raw_amount /= raw_target

    return raw_amount



"""
nanolib.accounts
~~~~~~~~~~~~~~~~

Methods for creating keys and IDs for NANO accounts and dealing with
wallet seeds
"""

from hashlib import blake2b

from collections import namedtuple

import binascii
import random
import re
import enum

from ed25519_blake2b import SigningKey

from .exceptions import (
    InvalidSeed, InvalidAccount, InvalidPublicKey, InvalidPrivateKey
)
from .util import (
    dec_to_hex, is_hex, nbase32_to_bytes, bytes_to_nbase32
)


__all__ = (
    "AccountKeyPair", "AccountIDPrefix", "validate_private_key",
    "validate_public_key", "is_account_id_valid", "validate_account_id",
    "validate_seed", "get_account_key_pair", "generate_account_private_key",
    "generate_account_key_pair", "get_account_id", "get_account_public_key",
    "generate_account_id", "generate_seed"
)


_AccountKeyPair = namedtuple("AccountKeyPair", ["private", "public"])


class AccountKeyPair(_AccountKeyPair):
    """NANO key pair

    :ivar str private: Private key as a 64-character hex string
    :ivar str public: Public key as a 64-character hex string
    """


class AccountIDPrefix(enum.Enum):
    """Prefixes used for NANO account IDs
    """

    XRB = "xrb_"
    """XRB prefix. Recognized universally."""

    NANO = "nano_"
    """NANO prefix. Newer prefix, may not be supported by all endpoints."""


ACCOUNT_ID_FIRST_DIGITS = ("1", "3")


def validate_private_key(private_key):
    """Validate the given private key and raise an exception on failure

    :param str private_key: Private key as a 64-character hex string
    :raises InvalidPrivateKey: If the private key is invalid
    :return: The private key
    :rtype: str
    """
    if len(private_key) != 64 or not is_hex(private_key):
        raise InvalidPrivateKey(
            "Account private key must be a 64-character hexadecimal string")

    return private_key


def validate_public_key(public_key):
    """Validate the given public key and raise an exception on failure

    :param str public_key: Public key as a 64-character hex string
    :raises InvalidPublicKey: If the public key is invalid
    :return: The public key
    :rtype: str
    """
    if len(public_key) != 64 or not is_hex(public_key):
        raise InvalidPublicKey(
            "Account public key must be a 64-character hexadecimal string")

    return public_key


def is_account_id_valid(account_id):
    """Check if the account ID is valid

    :param str account_id: Account ID as a string
    :return: True if the account ID is valid, False otherwise
    :rtype: bool
    """
    try:
        get_account_public_key(account_id=account_id)
        return True
    except InvalidAccount:
        return False


def validate_account_id(account_id):
    """Validate a NANO account ID and raise an exception on failure

    :param str account_id: Account ID as a string
    :raises InvalidAccount: If the account ID is invalid
    :return: The account ID
    :rtype: str
    """
    get_account_public_key(account_id=account_id)
    return account_id


def validate_seed(seed):
    """Validate a NANO seed and raise an exception on failure

    :param str seed: Seed as a 64-character hex string
    :raises InvalidSeed: If the seed is invalid
    :return: The seed
    :rtype: str
    """
    if len(seed) != 64 or not is_hex(seed):
        raise InvalidSeed("Seed must be a 64-character hexadecimal string")

    return seed


def get_account_key_pair(private_key):
    """Generate a public key from a private key and return the key pair

    :param private_key: Private key as a 64-character hex string
    :raises InvalidPrivateKey: If the private key is invalid
    :return: Return the key pair
    :rtype: AccountKeyPair
    """
    validate_private_key(private_key)

    signing_key = SigningKey(binascii.unhexlify(private_key))
    private_key = signing_key.to_bytes().hex()
    public_key = signing_key.get_verifying_key().to_bytes().hex()

    return AccountKeyPair(private=private_key, public=public_key)


def generate_account_private_key(seed, index):
    """Generate account's private key from a 32-byte seed and index

    :param str seed: Seed as a 64-character hex string
    :param int index: Index of the account private key to generate
    :raises InvalidSeed: If the seed is invalid
    :raises ValueError: If the index isn't an integer
    :return: Account private key as a 64-character hex string
    :rtype: str
    """
    validate_seed(seed)

    if not isinstance(index, int):
        raise ValueError("Index must be an integer")

    account_bytes = binascii.unhexlify(dec_to_hex(index, 4))
    context = blake2b(digest_size=32)
    context.update(binascii.unhexlify(seed))
    context.update(account_bytes)

    new_key = context.hexdigest()
    return new_key


def generate_account_key_pair(seed, index):
    """Generate an account key pair from a 32-byte seed and index

    :param str seed: Seed as a 64-character hex string
    :param int index: Index of the account key pair to generate
    :raises InvalidSeed: If the seed is invalid
    :raises ValueError: If the index isn't an integer
    :return: Account public and private key pair
    :rtype: AccountKeyPair
    """
    private_key = generate_account_private_key(seed, index)
    _, public_key = get_account_key_pair(private_key)

    return AccountKeyPair(private=private_key, public=public_key)


def get_account_id(*, public_key=None, private_key=None, prefix=None):
    """Get NANO account ID by using either a public key or a private key

    .. note:: Parameters `public_key` and `private_key` are mutually exclusive \
              and have to be given as keyword arguments

    :param str public_key: Public key as a 64-character hex string
    :param str private_key: Private key as a 64-character hex string
    :param str prefix: (optional) Prefix to use for the account ID. \
                       Can be either :const:`AccountIDPrefix.NANO` or \
                       :const:`AccountIDPrefix.XRB`. \
                       Defaults to :const:`AccountIDPrefix.XRB`

    :raises ValueError: If the given prefix isn't either XRB or NANO
    :raises TypeError: If both private and public keys are given, or if \
                       none are given as named parameter
    :raises InvalidPublicKey: If the given public key is invalid
    :raises InvalidPrivateKey: If the given private key is invalid
    :return: Account ID
    :rtype: str

     Usage::

      >>> from nanolib.accounts import get_account_id, AccountIDPrefix
      >>> get_account_id(private_key="2"*64)
      'xrb_1iwamgozb5ckj9zzojbnb79485dfiw8jegedzwzuzy5b4a19cbs8b4tsdzo3'
      >>> get_account_id(public_key="2"*64)
      'xrb_1aj46aj46aj46aj46aj46aj46aj46aj46aj46aj46aj46aj46aj4ykus34mi'
      >>> get_account_id(public_key="2"*64, prefix=AccountIDPrefix.NANO)
      'nano_1aj46aj46aj46aj46aj46aj46aj46aj46aj46aj46aj46aj46aj4ykus34mi'
      >>> get_account_id(public_key="2"*64, prefix="nano_")
      'nano_1aj46aj46aj46aj46aj46aj46aj46aj46aj46aj46aj46aj46aj4ykus34mi'

    """
    def _from_public_key(public_key, prefix):
        """
        Derive NANO account ID from a public key
        """
        validate_public_key(public_key)

        public_key_bytes = binascii.unhexlify(public_key)

        account = bytes_to_nbase32(public_key_bytes)
        checksum_bytes = blake2b(public_key_bytes, digest_size=5).digest()
        checksum_bytes = bytearray(checksum_bytes)
        checksum_bytes.reverse()
        checksum_bytes = bytes(checksum_bytes)
        checksum = bytes_to_nbase32(checksum_bytes)

        return "{prefix}{account}{checksum}".format(
            prefix=prefix, account=account, checksum=checksum)

    def _from_private_key(private_key, prefix):
        """
        Derive NANO account ID from a private key
        """
        _, public_key = get_account_key_pair(private_key=private_key)
        return _from_public_key(public_key=public_key, prefix=prefix)

    if not prefix:
        prefix = AccountIDPrefix.XRB
    try:
        prefix = AccountIDPrefix(prefix)
    except ValueError:
        raise ValueError(
            "Account ID prefix has to be one of the following: "
            "{prefixes}".format(
                prefixes=",".join(AccountIDPrefix._value2member_map_)
            )
        )

    if public_key and private_key:
        raise TypeError(
            "Only either 'public_key' or 'private_key' parameter can be given")
    elif not public_key and not private_key:
        raise TypeError(
            "Either 'public_key' or 'private_key' has to be given as a named "
            "parameter")

    if public_key:
        return _from_public_key(public_key, prefix.value)
    elif private_key:
        return _from_private_key(private_key, prefix.value)


def get_account_public_key(*, account_id=None, private_key=None):
    """Get a NANO public key using either an account ID or a private key

    .. note:: Parameters `account_id` and `private_key` are mutually exclusive \
              and have to be given as keyword arguments

    :param str account_id: Account ID
    :param str private_key: Private key as a 64-character hex string
    :raises TypeError: If both account ID and private key are given, or if \
                       none are given as keyword arguments
    :raises InvalidAccount: If the given account ID is invalid
    :raises InvalidPrivateKey: If the given private key is invalid
    :return: Public key as a 64-character hex string
    :rtype: str
    """

    def _from_account_id(account_id):
        try:
            account_prefix = account_id[0:account_id.index("_")+1]
            account_rest = account_id[account_id.index("_")+1:]
        except ValueError:
            raise InvalidAccount("Invalid NANO address")

        if account_prefix not in AccountIDPrefix._value2member_map_ or \
                account_rest[0] not in ACCOUNT_ID_FIRST_DIGITS or \
                len(account_rest) != 60:
            raise InvalidAccount("Invalid NANO address")

        try:
            account_bytes = nbase32_to_bytes(account_rest)
            public_key_bytes, account_checksum_bytes = \
                account_bytes[0:32], account_bytes[32:37]
        except ValueError:
            raise InvalidAccount("Invalid NANO address")

        key_checksum_bytes = bytearray(
            blake2b(public_key_bytes, digest_size=5).digest()
        )
        key_checksum_bytes.reverse()

        if key_checksum_bytes != account_checksum_bytes:
            raise InvalidAccount("Invalid checksum")

        return binascii.hexlify(public_key_bytes).decode()

    def _from_private_key(private_key):
        _, public_key = get_account_key_pair(private_key=private_key)

        return public_key

    if account_id and private_key:
        raise TypeError(
            "Only either 'account_id' or 'private_key' parameter can be given")
    elif not account_id and not private_key:
        raise TypeError(
            "Either 'account_id' or 'private_key' has to be given as a named "
            "parameter")

    if account_id:
        return _from_account_id(account_id)
    elif private_key:
        return _from_private_key(private_key)


def generate_account_id(seed, index):
    """Derive an account ID from a seed and an index

    :param str seed: Seed as a 64-character hex string
    :param int index: Index of the account ID to generate
    :return: Account ID
    :rtype: str
    """
    _, public_key = generate_account_key_pair(seed, index)
    account_id = get_account_id(public_key=public_key)

    return account_id


def generate_seed():
    """
    Generate a secure random 64-character hexadecimal seed for use in
    generating NANO accounts

    :return: Seed as a 64-character hex string
    :rtype: str
    """
    gen = random.SystemRandom()
    return "".join([gen.choice("0123456789abcdef") for _ in range(0, 64)])
